/**********************************************************************
 * Copyright (c) 2005 Scapa Technologies Limited and others
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * Scapa Technologies Limited - Initial API and implementation
 **********************************************************************/

package org.eclipse.stp.b2j.core.publicapi.engine;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Properties;

import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.model.IDebugTarget;
import org.eclipse.stp.b2j.core.B2jPlugin;
import org.eclipse.stp.b2j.core.jengine.internal.api.BpelProgramFactory;
import org.eclipse.stp.b2j.core.jengine.internal.api.EngineFactory;
import org.eclipse.stp.b2j.core.jengine.internal.api.Program;
import org.eclipse.stp.b2j.core.jengine.internal.compiler.Switches;
import org.eclipse.stp.b2j.core.jengine.internal.core.api.ControllerInterface;
import org.eclipse.stp.b2j.core.jengine.internal.core.api.DaemonInterface;
import org.eclipse.stp.b2j.core.jengine.internal.core.api.TraceListener;
import org.eclipse.stp.b2j.core.jengine.internal.message.Message;
import org.eclipse.stp.b2j.core.jengine.internal.utils.EngineListenerTraceListener;
import org.eclipse.stp.b2j.core.jengine.internal.utils.StreamUtils;
import org.eclipse.stp.b2j.core.jengine.internal.utils.TraceListenerTranslatorLog;
import org.eclipse.stp.b2j.core.publicapi.B2jPlatform;
import org.eclipse.stp.b2j.core.publicapi.DependencyInfo;
import org.eclipse.stp.b2j.core.publicapi.JARDependency;
import org.eclipse.stp.b2j.core.publicapi.importresolver.StandardWsdlImportResolver;
import org.eclipse.stp.b2j.core.publicapi.jcompiler.WorkbenchJDTJavaCompiler;
import org.eclipse.stp.b2j.core.publicapi.program.BPELProgram;
import org.eclipse.stp.b2j.core.publicapi.transport.session.SessionAddress;
import org.eclipse.stp.b2j.core.ui.internal.debug.ControllerDebugTarget;

/**
 * A public API which represents an engine capable of running a BPELProgram program
 * 
 * @author aem
 *
 */
public class WorkbenchBPELEngine {
	
	BPELProgram program;
	
	boolean compiled = false;
	
	Object running_LOCK = new Object();
	boolean running = false;
	
	ControllerInterface controller = null;
	Message root_runner_ids = null;

	IDebugTarget debugTarget = null;
	
	/**
	 * If the engine is running a program in debug mode then this returns a valid DebugTarget
	 * @return a valid DebugTarget, or null if the program is not running or is running in debug mode, or a launch has not been provided
	 */
	public IDebugTarget getDebugTarget() {
		return debugTarget;
	}
	
	/**
	 * Create a new BPEL engine to run a BPEL program
	 * @param program the BPEL program to run
	 */
	public WorkbenchBPELEngine(BPELProgram program) {
		this.program = program;
	}

	/**
	 * Run a BPEL program (will use the JVM local engine)
	 * @throws Exception if an error occurs while running the BPEL program
	 */
	public void runProgram() throws Exception {
		runProgram(new IgnoredTraceListener(),false,false,true,true);
	}

	/**
	 * Run a BPEL program
	 * @param useMiniEngine whether to use a local BPEL engine running in the current JVM
	 * @param terminateEngineOnDisconnect whether the engine should terminate if the client jvm disconnects
	 * @throws Exception if an error occurs while running the BPEL program
	 */
	public void runProgram(boolean useMiniEngine, boolean terminateEngineOnDisconnect) throws Exception {
		runProgram(new IgnoredTraceListener(),false,false,useMiniEngine,terminateEngineOnDisconnect);
	}
	
	/**
	 * Run a BPEL program
	 * @param debug whether to run the engine in debug mode (slower)
	 * @param verbose whether to produce verbose output when processing the input BPEL file etc.
	 * @param useMiniEngine whether to use a local BPEL engine running in the current JVM
	 * @param terminateEngineOnDisconnect whether the engine should terminate if the client jvm disconnects
	 * @throws Exception if an error occurs while running the BPEL program
	 */
	public void runProgram(boolean debug, boolean verbose, boolean useMiniEngine, boolean terminateEngineOnDisconnect) throws Exception {
		runProgram(new IgnoredTraceListener(),debug,verbose,useMiniEngine,terminateEngineOnDisconnect);
	}
	
	/**
	 * Run a BPEL program
	 * @param listener the trace listener to attach to the engine
	 * @param debug whether to run the engine in debug mode (slower)
	 * @param verbose whether to produce verbose output when processing the input BPEL file etc.
	 * @param useMiniEngine whether to use a local BPEL engine running in the current JVM
	 * @param terminateEngineOnDisconnect whether the engine should terminate if the client jvm disconnects
	 * @throws Exception if an error occurs while running the BPEL program
	 */
	public void runProgram(BPELEngineListener listener, boolean debug, boolean verbose, boolean useMiniEngine, boolean terminateEngineOnDisconnect) throws Exception {
		runProgram(listener,debug,verbose,useMiniEngine,terminateEngineOnDisconnect,null);
	}

	/**
	 * Run a BPEL program
	 * @param listener the trace listener to attach to the engine
	 * @param debug whether to run the engine in debug mode (slower)
	 * @param verbose whether to produce verbose output when processing the input BPEL file etc.
	 * @param useMiniEngine whether to use a local BPEL engine running in the current JVM
	 * @param terminateEngineOnDisconnect whether the engine should terminate if the client jvm disconnects
	 * @throws Exception if an error occurs while running the BPEL program
	 */
	public void runProgram(BPELEngineListener listener, boolean debug, boolean verbose, boolean useMiniEngine, boolean terminateEngineOnDisconnect, ILaunch launch) throws Exception {
		synchronized(running_LOCK) {
			runInternal(listener,debug,verbose,useMiniEngine,terminateEngineOnDisconnect,program,launch,false);
		}
		/*
		synchronized(running_LOCK) {
			
//			String bpel_file = program.bpel_file_path.toString();
//			String bpel_name = bpel_file.substring(Math.max(bpel_file.lastIndexOf('\\'),bpel_file.lastIndexOf('/')) + 1);
			SimpleDateFormat sdf = new SimpleDateFormat("yyyy_MM_dd HH:mm:ss.SSS");
			String engine_name = B2jPlugin.getString("B2J_RUN")+" ("+program.name+") ("+sdf.format(new Date(System.currentTimeMillis()))+")";
			BPELEngineListener console = new ConsoleTraceListener(engine_name);
			
			Program eprogram = BpelProgramFactory.createEngineProgramFromBpelSource(
					false,
					null,
					program.base_uri,
					program.source,
//					PathUtil.getLocationFromPath(program.bpel_file_path),
//					PathUtil.readFileFromPath(program.bpel_file_path),
					new JARDependency[0],
					null
					);		
			
			for (int i = 0; i < program.workers.size(); i++) {
				eprogram.addHostAddress((SessionAddress)program.workers.get(i),(SessionAddress)program.worker_daemons.get(i));
			}
	
			DaemonInterface daemon;
			
			if (program.workers.size() == 0) {
				//mini engine
				daemon = EngineFactory.connectToMiniEngineDaemon();

				controller = daemon.newEngine(engine_name,console,new JARDependency[0],null);
			} else {
				//main engine
	
				URL plugin_dir = Platform.asLocalURL(Platform.find(B2jPlugin.getDefault().getBundle(),new Path("/")));
				File plugin_file = new File(plugin_dir.getFile());
				File jar_file = new File(plugin_dir.getFile()+"/b2j.jar");
				
				EngineFactory.createMainEngineDaemon(plugin_file,jar_file,11000);
	
				daemon = EngineFactory.connectToMainEngineDaemon(program.coordinator_daemon);

				controller = daemon.newEngine(engine_name,console,new JARDependency[0],program.coordinator);
			}
			
			
			//dont need the daemon any more
			daemon.close();
			
			controller.setProgram(eprogram);
	
			controller.setLogLevel(true,true,false);
	
			root_runner_ids = controller.launchRunner(1,"engine_main",0,new String[0]);
			
			running = true;
		}*/
	}
	
	/**
	 * Forcibly terminate this BPEL program
	 * @throws Exception if an error occurs while waiting for this BPEL program to terminate
	 */
	public void terminate() throws Exception {
		controller.terminate();
	}
	
	/**
	 * Wait for the BPEL program to terminate
	 * @throws Exception if an error occurs while waiting for this BPEL program to terminate
	 */
	public void waitForProgramCompletion() throws Exception {
		
		synchronized(running_LOCK) {
			if (!running) {
				return;
			}
		
			for (int i = 0; i < root_runner_ids.length(); i++) {
				Long id = (Long)root_runner_ids.get(i);	
	//			try {
					controller.joinRunner(id);
	//			} catch (Exception e) {
	//				console.debug("Failed to wait on program finish");
	//				console.debug(getStacktrace(e));
	//				B2jPlugin.DBG.logVisibleError(e,"Failed to wait on program finish",false);
	//				return;
	//			}
			}
	
			//give the data reading and printing thread a chance to finish reading any stuff
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {}
	
			//close the connection to the engine
			controller.closeConnection();
			
			running = false;
			
			debugTarget = null;
		}
	}

	/**
	 * Compile and validate this BPEL program
	 * @return
	 * @throws Exception
	 */
	public BPELProgram compileProgram() throws Exception {
		compileProgram(new IgnoredTraceListener(),program,false,false,false,true);
		return program;
	}
	
	public void markProgramCompiled() throws Exception {
		compiled = true;
	}
	
	/**
	 * Validate this BPEL program, throws an exception if the validation fails for any reason
	 * @throws Exception
	 */
	public void validateProgram() throws Exception {
		compileProgram(new IgnoredTraceListener(),program,false,false,true,true);
//		runInternal(new IgnoredTraceListener(),false,false,true,true,program,null,true);
	}
	
	
	private void compileProgram(BPELEngineListener console, BPELProgram bpel_program, boolean debug, boolean verbose, boolean validateOnly, boolean uniquePackages) throws Exception {

		TraceListener console_tracel = new EngineListenerTraceListener(console);
		
//		long tott = System.currentTimeMillis();
		
		ArrayList program_deps = fetchAllProgramDependencies(console,bpel_program);
		
		//
		//Create the engine program
		//NOTE: we do this BEFORE we remove Program dependancies that are also Engine dependancies,
		//otherwise we miss out on 
		//
		Program prog = null;
		try {
			JARDependency[] tmp_program_deps = dependencyListToArray(program_deps);

			long t = System.currentTimeMillis();

			console.printInfo("Generating engine program...");

			Switches switches = new Switches();
			
			if (debug) {
				//make everything global so we can access it
				switches.FLOW_GRAPH_ASSET_OPTIMISATION = false;
				switches.SINGLE_HOST_ASSET_OPTIMISATION = false;
				//leave all the proper names etc in
				switches.MAP_XSDTYPE_TO_INDEXMAP = false;
				switches.RICH_ENGINE_SERIALISATION = true;
			}
			
			//
			// Set debugging compilation switches
			//
/*			if (mode.equalsIgnoreCase("debug")) {
				Switches.APPEND_TRANSACTIONS_TO_CALLSTACK = true;
			} else {
				Switches.APPEND_TRANSACTIONS_TO_CALLSTACK = false;
			}
*/			
			console.printInfo("");
			TraceListenerTranslatorLog mclog = new TraceListenerTranslatorLog(console_tracel,verbose);
			
			prog = BpelProgramFactory.createEngineProgramFromBpelSource(debug,switches,new WorkbenchJDTJavaCompiler(),new StandardWsdlImportResolver(),bpel_program.getBaseURI(),bpel_program.getBpelSource(),tmp_program_deps,mclog,validateOnly,uniquePackages);

			console.printInfo("");
			t = System.currentTimeMillis()-t;
			t /=1000;
			console.printInfo("Generated engine program ("+t+"sec)");
			console.printInfo("");
			
			bpel_program.setCompiledData("Program",prog);
			bpel_program.setCompiledData("Program Dependencies",program_deps);
			
			compiled = true;
			
		} catch (Exception e) {
			if (validateOnly) throw e;
			
			printFatalProblem(console,"Failed to build engine program from BPEL source",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to build engine program from BPEL source",false);
			
			return;
		}
	}
	
	private void runInternal(BPELEngineListener console, boolean debug, boolean verbose, boolean useMiniEngine, boolean exitOnDisconnect, BPELProgram bpel_program, ILaunch launch, boolean validateOnly) throws Exception {

		TraceListener console_tracel = new EngineListenerTraceListener(console);
		
//		System.out.println(bpel_program.source);

		long tott = System.currentTimeMillis();
		
//		ArrayList program_deps = fetchAllProgramDependencies(console);
		
		if (!compiled) {
			compileProgram(console,bpel_program,debug,verbose,false,useMiniEngine);
		}
		
		Program prog = (Program)bpel_program.getCompiledData("Program");
		ArrayList program_deps = (ArrayList)bpel_program.getCompiledData("Program Dependencies");
		
		String[] deps = bpel_program.getProgramDependencies();
		
//		console.printInfo("Compiled program dependency count: "+program_deps.size());
//		console.printInfo("Uncompiled program dependency count: "+deps.length);
//		if (program_deps.size() != deps.length) {
//			console.printDebug("WARNING: program may have been modified post-compilation?");
//		}
		
		SessionAddress[] workers = bpel_program.getWorkerHosts();
		SessionAddress[] worker_daemons = bpel_program.getWorkerDaemons();
		
		for (int i = 0; i < workers.length; i++) {
			prog.addHostAddress(workers[i],worker_daemons[i]);
		}

		ArrayList engine_deps = fetchAllEngineDependencies(console);
		
		//
		//remove any multiples from Program deps that are included in Engine deps
		//
		for (int i = 0; i < program_deps.size(); i++) {
			JARDependency dep = (JARDependency)program_deps.get(i);
			for (int k = 0; k < engine_deps.size(); k++) {
				JARDependency tmp = (JARDependency)engine_deps.get(k);
				if (dep.getFilePath().equals(tmp.getFilePath())) {
					Object o = program_deps.remove(i--);
					console.printInfo("Removing duplicate dependency: "+o);
					break;
				}
			}
		}
		
		JARDependency[] engine_deps_array = dependencyListToArray(engine_deps);
		JARDependency[] program_deps_array = dependencyListToArray(program_deps);
		
		if (engine_deps_array.length > 0 || program_deps_array.length > 0) {
			console.printInfo("Dependencies:");
		}

		for (int i = 0; i < engine_deps_array.length; i++) {
			console.printInfo("  (Engine Dependency)  "+engine_deps_array[i].getFilePath());
		}
		for (int i = 0; i < program_deps_array.length; i++) {
			console.printInfo("  (Program Dependency)  "+program_deps_array[i].getFilePath());
		}

		if (engine_deps_array.length > 0 || program_deps_array.length > 0) {
			console.printInfo("");
		}
		
		String coordinator_hostname = null;
		
		DaemonInterface daemon = null;
		try {
			console.printInfo("Connecting to engine daemon...");
			long t = System.currentTimeMillis();
			if (!useMiniEngine) {

				URL plugin_dir = Platform.asLocalURL(Platform.find(B2jPlugin.getDefault().getBundle(),new Path("/")));
				File plugin_file = new File(plugin_dir.getFile());
				File jar_file = new File(plugin_dir.getFile()+"/b2j.jar");
				
				//start an engine daemon 
				console.printInfo("Starting main engine daemon in "+plugin_file);
				EngineFactory.createMainEngineDaemon(plugin_file,jar_file,11000);
	
				SessionAddress daemon_address = bpel_program.getCoordinatorDaemon();
//				SessionAddress daemon_address = (SessionAddress)daemon_list.get(0);
				coordinator_hostname = daemon_address.getListenerHost();

				String protocol;
				if (daemon_address.getRequiresEncryption()) {
					protocol = "(HTTPS)";
				} else {
					protocol = "(HTTP)";
				}
				
				String password;
				if (daemon_address.getRequiresPassword()) {
					password = "(requires password)";
				} else {
					password = "(no password)";
				}
				
				//connect to the engine daemon 
				console.printInfo("Connecting to main engine daemon on "+coordinator_hostname+" "+protocol+" "+password);
				daemon = EngineFactory.connectToMainEngineDaemon(daemon_address);
				
			} else {
				//connect to the engine daemon 
				console.printInfo("Connecting to in-jvm mini engine daemon");
				daemon = EngineFactory.connectToMiniEngineDaemon();
				
			}
			
			t = System.currentTimeMillis()-t;
			t /=1000;
			console.printInfo("Connected to engine daemon ("+t+"sec)");
			console.printInfo("");
		} catch (Exception e) {
			printFatalProblem(console,"Failed to connect to engine daemon",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to connect to engine daemon",false);
			return;
		}
		
		//start up the engine
		controller = null;
		try {
			long t = System.currentTimeMillis();
			console.printInfo("Creating engine instance...");
			controller = daemon.newEngine(bpel_program.getName(),console_tracel,engine_deps_array,bpel_program.getCoordinatorHost());
			
			//set JAR dependencies
			
			daemon.close();
			t = System.currentTimeMillis()-t;
			t /=1000;
			console.printInfo("Created engine instance ("+t+"sec)");
			console.printInfo("");
		} catch (Exception e) {
			printFatalProblem(console,"Failed to create new engine instance",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to create new engine instance",false);
			return;
		}
		
		long offset = 0;
		try {
			long t = System.currentTimeMillis();
			
			console.printInfo("Setting engine program...");

			if (!useMiniEngine) {
				console.printInfo("");
				console.printInfo("Distribution:");
				console.printInfo("  (Coordinator host)   "+coordinator_hostname);
				
				for (int i = 0; i < prog.getAddressCount(); i++) {
					console.printInfo("  (Worker host)        "+((SessionAddress)prog.getAddresses().get(i)).getListenerHost());
				}
				console.printInfo("");
			}
			
			offset = controller.setProgram(prog);
			
			t = System.currentTimeMillis()-t;
			t /=1000;
			console.printInfo("Set engine program ("+t+"sec)");
			console.printInfo("");
		} catch (Exception e) {
			printFatalProblem(console,"Failed to set engine program in engine instance",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to set engine program in engine instance",false);
			return;
		}

		try {
			if (!exitOnDisconnect) {
				controller.setHeadless(true);
				console.printInfo("Headless mode, engine will continue to run after workbench disconnects");
				console.printInfo("");
			}
		} catch (Exception e) {
			printFatalProblem(console,"Failed to set headless in engine instance",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to set headless in engine instance",false);
			return;
		}
		
		try {
			if (debug) {
				console.printInfo("Debug mode, turning on all engine logging");
				console.printInfo("");
				controller.setLogLevel(true,true,false);
//				controller.setLogLevel(true,true,true);
			} else {
				console.printInfo("Turning on only warning and error logging");
				console.printInfo("");
				controller.setLogLevel(true,true,false);
			}
		} catch (Exception e) {
			printFatalProblem(console,"Failed to set log level in engine instance",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to set log level in engine instance",false);
			return;
		}

		//
		// NB: do this before we launch the runner - it initialises the variables and hashmaps etc
		//
		if (debug) {
			if (launch != null) {
				debugTarget = new ControllerDebugTarget(launch,controller,bpel_program.getName());
			}
		}

		tott = System.currentTimeMillis()-tott;
		tott /=1000;
		
		console.printInfo("Total launch time "+tott+"sec");
		console.printInfo("");
		
		try {
			console.printInfo("Launching engine program");
			console.printInfo("");
			root_runner_ids = controller.launchRunner(1,"engine_main",0,new String[0]);
		} catch (Exception e) {
			printFatalProblem(console,"Failed to start engine program",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to start engine program",false);
			return;
		}
		
		running = true;
		
/*
		Message root_runner_ids = null;
		try {
			console.printInfo("Launching engine program");
			console.printInfo("");
			root_runner_ids = controller.launchRunner(1,"engine_main",0,new String[0]);
		} catch (Exception e) {
			printFatalProblem(console,"Failed to start engine program",e);
			B2jPlugin.DBG.logVisibleError(e,"Failed to start engine program",false);
			return;
		}

		AbortTracker tracker = new AbortTracker(controller,monitor);
		tracker.start();

		for (int i = 0; i < root_runner_ids.length(); i++) {
			Long id = (Long)root_runner_ids.get(i);	
			try {
				controller.joinRunner(id);
			} catch (Exception e) {
				printFatalProblem(console,"Failed to wait on program finish",e);
				B2jPlugin.DBG.logVisibleError(e,"Failed to wait on program finish",false);
				return;
			}
		}

		tracker.finished();

		try {
			Thread.sleep(1000);
		} catch (Exception e) {}
		
		//give the data reading and printing thread a chance to finish reading any stuff
//		try {
//			Thread.sleep(8000);
//		} catch (Exception e) {}

		//close the connection to the engine
		controller.closeConnection();

		//give the data reading and printing thread a chance to finish reading any stuff
//		try {
//			Thread.sleep(8000);
//		} catch (Exception e) {}*/
	}
	
	//
	// Utility methods
	//
	
	/**
	 * Convert a list of JARDependency objects to an array
	 * @param list
	 * @return
	 */
	private JARDependency[] dependencyListToArray(ArrayList list) {
		JARDependency[] array = new JARDependency[list.size()];
		list.toArray(array);
		return array;
	}
		
	private void printFatalProblem(BPELEngineListener console, String description, Exception e) {
		console.printDebug("\n\nERROR: "+description);
		console.printInfo("\nProblem Description:");
		if (e.getMessage() != null) {
			console.printDebug(e.getMessage());
		} else {
			console.printDebug("");
		}
		console.printInfo("\nProblem Details:");
		console.printDebug(getStacktrace(e));
	}
	
	private static String getStacktrace(Throwable t) {
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		PrintStream print = new PrintStream(bout);
		t.printStackTrace(print);
		return new String(bout.toByteArray()).replace('\t',' ');
	}



	/**
	 * Fetch all the JAR dependencies which the engine program has (e.g. user specified JARs or 
	 * WSDL binding JARs etc)
	 * @param console
	 * @return
	 */
	private ArrayList fetchAllProgramDependencies(BPELEngineListener console, BPELProgram prog) throws IOException {
		DependencyInfo[] infos = B2jPlatform.getPortDependencyInfo();
		return fetchAllDependencies(console,infos,prog);
	}
	
	/**
	 * Fetch all the JAR dependencies which the underlying java engine has (e.g. transport
	 * extension point JARs etc)
	 * @param console
	 * @return
	 */
	private ArrayList fetchAllEngineDependencies(BPELEngineListener console) throws IOException {
		DependencyInfo[] infos = B2jPlatform.getEngineDependencyInfo();
		return fetchAllDependencies(console,infos,null);
	}
	
	private ArrayList fetchAllDependencies(BPELEngineListener console, DependencyInfo[] infos, BPELProgram prog) throws IOException {
		ArrayList jardeps = new ArrayList();

		if (prog != null) {
			String[] paths = prog.getProgramDependencies();
			for (int k = 0; k < paths.length; k++) {
				
				InputStream in = new FileInputStream(paths[k]);
				in = new BufferedInputStream(in);
				
				byte[] dat = StreamUtils.readAll(in);
				
				in.close();
				
				if (paths[k] != null) {
					JARDependency dep = new JARDependency(dat,paths[k]);
					jardeps.add(dep);
				}
			}
		}
		
		
		for (int i = 0; i < infos.length; i++) {
			Properties[] props = infos[i].getResources();
			for (int k = 0; k < props.length; k++) {
				String path = props[k].getProperty("JAR");
				
				path = infos[i].getRelativePath(path);
				
				InputStream in = new FileInputStream(path);
				in = new BufferedInputStream(in);
				
				byte[] dat = StreamUtils.readAll(in);
				
				in.close();
				
				if (path != null) {
					JARDependency dep = new JARDependency(dat,path);
					jardeps.add(dep);
				}
			}
		}
		return jardeps;
	}
	/**
	 * Fetch all the JAR dependencies which the underlying java engine has (e.g. transport
	 * extension point JARs etc)
	 * @param console
	 * @return
	 */
/*	private ArrayList fetchAllEngineDependencies(BPELEngineListener console) {
		ArrayList engine_dependants = new ArrayList(); 
		
		try {
			Object[] deps = SessionTransportLoader.getTransports();
			for (int i = 0; i < deps.length; i++) {
				engine_dependants.add((Dependant)deps[i]);
			}
		} catch (Exception e) {
			console.printDebug("Failed to fetch JAR dependencies for Session Transport extensions");
			console.printDebug(getStacktrace(e));
			B2jPlugin.DBG.logVisibleError(e,"Failed to fetch JAR dependencies for Session Transport extensions",false);
			return null;
		}
			
		ArrayList engine_dependancies = new ArrayList();

		try {
			engine_dependancies = dependantsToDependencies("Engine (transport)",console,engine_dependants);
		} catch (Exception e) {
			console.printDebug("Failed to fetch JAR dependency");
			console.printDebug(getStacktrace(e));
			B2jPlugin.DBG.logVisibleError(e,"Failed to fetch JAR dependency",false);
			return null;
		}
		
		engine_dependancies = filterMultiples(engine_dependancies);
		
		return engine_dependancies;
	}
	*/
	/**
	 * Fetch all the JAR dependencies which the engine program has (e.g. user specified JARs or 
	 * WSDL binding JARs etc)
	 * @param console
	 * @return
	 */
/*	private ArrayList fetchAllProgramDependencies(BPELEngineListener console, BPELProgram bpel_program, ILaunchConfiguration conf) {
		ArrayList program_dependants = new ArrayList(); 
		
		try {
			Object[] deps = WSDLBindingLoader.getBindings();
			for (int i = 0; i < deps.length; i++) {
				program_dependants.add((Dependant)deps[i]);
			}
		} catch (Exception e) {
			console.printDebug("Failed to fetch JAR dependencies for WSDL Binding extensions");
			console.printDebug(getStacktrace(e));
			B2jPlugin.DBG.logVisibleError(e,"Failed to fetch JAR dependencies for WSDL Binding extensions",false);
			return null;
		}

		ArrayList program_dependencies = new ArrayList();
		try {
			program_dependencies = dependantsToDependencies("Program (wsdl binding)",console,program_dependants);
			
			String[] jar_deps = bpel_program.getProgramDependencies();
			for (int i = 0; i < jar_deps.length; i++) {
				JARDependency tmp = new JARDependency(new byte[0],jar_deps[i]);
				tmp.setEnabled(true);
				program_dependencies.add(tmp);
			}
			
			//
			// Load extra Program dependencies from depedency extension point
			//
			JARDependency[] exts = DependencyLoader.getDependencies();
			for (int i = 0; i < exts.length; i++) {

				boolean enabled = true;
//TODO				boolean enabled = conf.getAttribute("jar_enabled_"+B2jDependenciesTab.getExtensionDependencyKey(exts[i]),true);
				exts[i].setEnabled(enabled);
				
				if (exts[i].getEnabled()) {
					//
					//Read the file in from the plugin
					//
					Bundle bundle = Platform.getBundle(exts[i].getPluginName());
					URL plugin_dir = Platform.asLocalURL(Platform.find(bundle,new Path(exts[i].getFilePath())));
					File jar_file = new File(plugin_dir.getFile());
					String cpath = jar_file.getCanonicalPath();
		
					InputStream in = new FileInputStream(cpath);
					byte[] dat = StreamUtils.readAll(in);
					in.close();
					
					JARDependency dependency = new JARDependency(dat,cpath);

					program_dependencies.add(dependency);
				}
			}
			
		} catch (Exception e) {
			console.printDebug(e.toString());
			B2jPlugin.DBG.logVisibleError(e,e.toString(),false);
			return null;
		}
		
		//
		//filter out multiples from engine dependencies and 
		//
		program_dependencies = filterMultiples(program_dependencies);
	
		return program_dependencies;
	}*/
		
	/**
	 * Take a list of JARDependant objects and load all the JARDependency objects they
	 * say they require
	 * @param msg_prefix
	 * @param console
	 * @param dependants
	 * @return
	 * @throws Exception
	 *//*
	private ArrayList dependantsToDependencies(String msg_prefix, BPELEngineListener console, ArrayList dependants) throws Exception {
		ArrayList jar_deps = new ArrayList();
		for (int i = 0; i < dependants.size(); i++) {
			Dependant dep = (Dependant)dependants.get(i);
			Bundle bundle = Platform.getBundle(dep.getProviderPluginName());
			
			String[] paths = dep.getJarDependencyProvider().getJarDependencies();
			
			for (int k = 0; k < paths.length; k++) {
				try {
					URL plugin_dir = Platform.asLocalURL(Platform.find(bundle,new Path(paths[k])));
					File jar_file = new File(plugin_dir.getFile());
					String cpath = jar_file.getCanonicalPath();
		
					InputStream in = new FileInputStream(cpath);
					byte[] dat = StreamUtils.readAll(in);
					in.close();
					
					JARDependency dependency = new JARDependency(dat,cpath);
					jar_deps.add(dependency);
				} catch (Exception e) {
					e = new Exception("Failed to get JAR dependency ("+paths[k]+"):"+e);
					throw e;
				}
			}
		}
		
		return jar_deps;
	}*/
	
	/**
	 * Check to see if there are multiples of JAR dependencies and remove any multiples
	 * @param deps
	 * @return
	 *//*
	private ArrayList filterMultiples(ArrayList deps) {
		ArrayList uniques = new ArrayList();
		for (int i = 0; i < deps.size(); i++) {
			JARDependency dep = (JARDependency)deps.get(i);
			
			boolean unique = true;
			for (int k = 0; k < uniques.size(); k++) {
				JARDependency cmp = (JARDependency)uniques.get(k);
				if (dep.getFilePath().equals(cmp.getFilePath())) {
					unique = false;
					break;
				}
			}
			
			if (unique) uniques.add(dep);
		}
		return uniques;
	}
	*/
	private class IgnoredTraceListener implements BPELEngineListener {

		public void printInfo(String s) {
		}

		public void printDebug(String s) {
		}

		public void printEngineInfo(String s) {
		}

		public void printEngineDebug(String s) {
		}
	}
}